#include "App.h"

BaseApp *app = new App();

void App::onButtonClicked(PushButton *button){
	if (button == bench){
		isBenching = true;
		nFrames = 0;
		if (complex->isChecked()){
			maxFrames = 512;
		} else {
			maxFrames = 4096;
		}
		configDialog->setVisible(false);
	} else {
		BaseApp::onButtonClicked(button);
	}
}

bool App::init(){
	int tab = configDialog->addTab("Tests");

	bench = new PushButton(0, 128, 192, 36, "Run benchmark");
	bench->setListener(this);
	configDialog->addWidget(tab, bench);

	overDrawMode = new DropDownList(0, 10, 270, 38);
	overDrawMode->addItem("Front to back");
	overDrawMode->addItem("Back to front");
	overDrawMode->addItem("Random order");
	overDrawMode->addItem("Random with pre-Z");
	overDrawMode->selectItem(0);
	configDialog->addWidget(tab, overDrawMode);

	overDrawFactor = new DropDownList(280, 10, 88, 38);
	overDrawFactor->addItem("2x");
	overDrawFactor->addItem("3x");
	overDrawFactor->addItem("4x");
	overDrawFactor->addItem("8x");
	overDrawFactor->addItem("16x");
	overDrawFactor->addItem("64x");
	overDrawFactor->selectItem(3);
	configDialog->addWidget(tab, overDrawFactor);

	complex = new CheckBox(0, 60, 220, 38, "Complex shader");
	configDialog->addWidget(tab, complex);


	configDialog->setCurrentTab(tab);
	return true;
}


bool App::load(){
	if (!GL_ARB_fragment_program_supported){
		WarningMsg("GL_ARB_fragment_program not supported. Complex shader mode disabled.");
		complex->setVisible(false);
	}

	if ((trilinear = renderer->addSamplerState(TRILINEAR, WRAP, WRAP, WRAP, 0.5f)) == SS_NONE) return false;

	if ((wall = renderer->addTexture("../Textures/pipetech.dds", true, trilinear)) == SHADER_NONE) return false;
//	if ((wall = renderer->addTexture("../Textures/pipetech.dds", TRILINEAR, 0, 0.5f)) == SHADER_NONE) return false;

	if ((less = renderer->addDepthState(true, true, LESS)) == BS_NONE) return false;
	if ((noColor = renderer->addBlendState(ONE, ZERO, BM_ADD, 0)) == BS_NONE) return false;

	configDialog->setVisible(true);
	isBenching = false;
	captureMouse(false);

	char fsText[4096];

	FILE *file = fopen("Shader.fp", "rb");
	if (file == NULL) return false;
	size_t n = fread(fsText, 1, sizeof(fsText), file);
	fsText[n] = '\0';
	fclose(file);

	glGenProgramsARB(1, &fp);
	glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, fp);
	glGetError();
	glProgramStringARB(GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, (GLsizei) strlen(fsText), fsText);
	GLint error = glGetError();
	if (error != GL_NO_ERROR){
		ErrorMsg((char *) glGetString(GL_PROGRAM_ERROR_STRING_ARB));
		return false;
	}

	return true;
}

void App::unload(){
	glDeleteProgramsARB(1, &fp);
}

void App::drawODQuad(const float z, const float t){
	glBegin(GL_TRIANGLE_FAN);
		glTexCoord4f(-1, -1, 0, t);
		glVertex3f(0, 0, z);

		glTexCoord4f( 1, -1, 0, t);
		glVertex3f(1, 0, z);

		glTexCoord4f( 1,  1, 0, 1.0f / t);
		glVertex3f(1, 1, z);

		glTexCoord4f(-1,  1, 0, 1.0f / t);
		glVertex3f(0, 1, z);
	glEnd();
}

float App::testOverDraw(const int overdrawFactor, const DrawMode mode, const bool complex){
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, 1, 1, 0, 1, -1);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

	// Clear states
	renderer->reset();
	renderer->apply();

	glClear(GL_DEPTH_BUFFER_BIT);


	float t = 0, z = 0, f = nFrames * 10.0f / maxFrames;
	float zr[MAX_OVERDRAW];
	if (mode >= RandomOrder){
		for (int i = 0; i < overdrawFactor; i++){
			zr[i] = float(rand() & 32767) / 32768.0f;
		}
		t = float(0.291667f * sinf(f) + 1.0416667f);

		if (mode >= RandomWithPreZ){
			renderer->reset();
			renderer->setDepthState(less);
			renderer->setBlendState(noColor);
			renderer->apply();

			for (int i = 0; i < overdrawFactor; i++){
				drawODQuad(zr[i], t);
			}
		}
	}

	renderer->reset();
	if (complex){
		glEnable(GL_FRAGMENT_PROGRAM_ARB);
		glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, fp);
	} else {
		((OpenGLRenderer *) renderer)->setTexture(wall);
	}
	renderer->apply();

	for (int i = 0; i < overdrawFactor; i++){
		switch (mode){
		case FrontToBack:
			t = float(0.291667f * sinf(f + i) + 1.0416667f);
			z = float(i) / overdrawFactor;
			break;
		case BackToFront:
			t = float(0.291667f * sinf(f + i) + 1.0416667f);
			z = -float(i) / overdrawFactor;
			break;
		case RandomOrder:
		case RandomWithPreZ:
			z = zr[i];
			break;
		}

		drawODQuad(z, t);
	}

	if (complex){
		glDisable(GL_FRAGMENT_PROGRAM_ARB);
	}

	return 0;
}

void App::drawFrame(){
	if (isBenching){
		static int odFactor[] = { 2, 3, 4, 8, 16, 64 };

		if (nFrames == 0){
			glFinish();
			updateTime();
			startTime = time;
		}

		if (nFrames == maxFrames){
			glFinish();
			updateTime();
			float fps = maxFrames / (time - startTime);

			static char *names[] = { "Front to back", "Back to front", "Random order", "Random order with pre-Z" };
			char str[256];
			sprintf(str, "%s - %dx overdraw - %s\n%.2f fps\n\n", complex->isChecked()? "Complex" : "Simple",
				odFactor[overDrawFactor->getSelectedItem()], names[overDrawMode->getSelectedItem()], fps);
			InfoMsg(str);

			FILE *file = fopen("result.log", "a");
			fwrite(str, 1, strlen(str), file);
			fclose(file);

			isBenching = false;
			configDialog->setVisible(true);
		} else {
			testOverDraw(odFactor[overDrawFactor->getSelectedItem()],
				(DrawMode) overDrawMode->getSelectedItem(), complex->isChecked());

			nFrames++;
		}


	} else {
		// Reset states
		renderer->reset();
		renderer->apply();
		glClearColor(0.2f, 0.4f, 0.3f, 0);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	}
}
